/*
 * MSC (MMC/SD controller) peripheral support.
 *
 * Copyright (C) 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <l4/devices/msc-common.h>



// Register locations for each channel.

enum Regs : unsigned
{
  Msc_control                       = 0x000,      // MSC_CTRL
  Msc_status                        = 0x004,      // MSC_STAT
  Msc_clock_rate                    = 0x008,      // MSC_CLKRT
  Msc_command_data_control          = 0x00c,      // MSC_CMDAT
  Msc_response_timeout              = 0x010,      // MSC_RESTO
  Msc_read_timeout                  = 0x014,      // MSC_RDTO
  Msc_block_length                  = 0x018,      // MSC_BLKLEN
  Msc_block_count                   = 0x01c,      // MSC_NOB
  Msc_block_success_count           = 0x020,      // MSC_SNOB
  Msc_interrupt_mask                = 0x024,      // MSC_IMASK
  Msc_interrupt_flag                = 0x028,      // MSC_IFLG/MSC_IREG
  Msc_command_index                 = 0x02c,      // MSC_CMD
  Msc_command_argument              = 0x030,      // MSC_ARG
  Msc_response_fifo                 = 0x034,      // MSC_RES
  Msc_recv_data_fifo                = 0x038,      // MSC_RXFIFO
  Msc_trans_data_fifo               = 0x03c,      // MSC_TXFIFO

  // JZ4780/X1600 only.

  Msc_low_power_mode                = 0x040,      // MSC_LPM
  Msc_dma_control                   = 0x044,      // MSC_DMAC
  Msc_dma_descriptor_address        = 0x048,      // MSC_DMANDA
  Msc_dma_data_address              = 0x04c,      // MSC_DMADA
  Msc_dma_data_length               = 0x050,      // MSC_DMALEN
  Msc_dma_command                   = 0x054,      // MSC_DMACMD
  Msc_control2                      = 0x058,      // MSC_CTRL2
  Msc_rtfifo_data_counter           = 0x05c,      // MSC_RTCNT

  // Channel block size/offset.

  Msc_channel_offset                = 0x10000,
};

// Field definitions.

enum Control_bits : unsigned
{
  // JZ4780/X1600 only.

  Control_send_ccsd                 = 0x8000,     // SEND_CCSD
  Control_send_ccsd_automatically   = 0x4000,     // SEND_CCSD

  // Common.

  Control_exit_multiple             = 0x0080,     // EXIT_MULTIPLE
  Control_exit_transfer             = 0x0040,     // EXIT_TRANSFER
  Control_start_read_wait           = 0x0020,     // START_READ_WAIT
  Control_stop_read_wait            = 0x0010,     // STOP_READ_WAIT
  Control_reset                     = 0x0008,     // RESET
  Control_start_operation           = 0x0004,     // START_OP

  Control_clock_control_field_mask  = 0x3,        // CLOCK_CTRL
  Control_clock_control_start       = 2,
  Control_clock_control_stop        = 1,
  Control_clock_control_field_shift = 0,
};

enum Control2_bits : unsigned
{
  // JZ4780/X1600 only.

  Control2_pin_level_polarity_field_mask  = 0x1f, // PIP
  Control2_pin_level_polarity_field_shift = 24,

  // JZ4780 only.

  Control2_reset_enable             = 0x00800000, // RST_EN

  // JZ4780/X1600 only.

  Control2_stop_read_operation_mode = 0x00000010, // STPRM

  // JZ4780 only.

  Control2_signal_voltage_change    = 0x00000008, // SVC

  // JZ4780/X1600 only.

  Control2_speed_mode_field_mask    = 0x7,        // SMS
  Control2_speed_mode_default       = 0,          // = 0
  Control2_speed_mode_high          = 1,          // = 1
  Control2_speed_mode_sdr12         = 2,          // = 2
  Control2_speed_mode_sdr25         = 3,          // = 3
  Control2_speed_mode_sdr50         = 4,          // = 4
  Control2_speed_mode_field_shift   = 0,
};

enum Status_bits : unsigned
{
  // JZ4780/X1600 only.

  Status_auto_cmd12_done            = 0x80000000, // AUTO_CMD12_DONE

  // JZ4780 only.

  Status_auto_cmd23_done            = 0x40000000, // AUTO_CMD23_DONE
  Status_signal_voltage_change      = 0x20000000, // SVS

  // JZ4780/X1600 only.

  Status_pin_level_field_mask       = 0x1f,       // PIN_LEVEL
  Status_pin_level_field_shift      = 24,

  Status_boot_crc_error             = 0x00100000, // BCE
  Status_boot_data_end              = 0x00080000, // BDE
  Status_boot_ack_error             = 0x00040000, // BAE
  Status_boot_ack_received          = 0x00020000, // BAR
  Status_dma_end                    = 0x00010000, // DMAEND

  // Common.

  Status_resetting                  = 0x8000,     // IS_RESETTING
  Status_sdio_interrupt_active      = 0x4000,     // SDIO_INT_ACTIVE
  Status_programming_done           = 0x2000,     // PRG_DONE
  Status_data_transfer_done         = 0x1000,     // DATA_TRAN_DONE
  Status_end_command_response       = 0x0800,     // END_CMD_RES
  Status_data_fifo_almost_full      = 0x0400,     // DATA_FIFO_AFULL
  Status_read_wait                  = 0x0200,     // IS_READWAIT
  Status_clock_enabled              = 0x0100,     // CLK_EN
  Status_data_fifo_full             = 0x0080,     // DATA_FIFO_FULL
  Status_data_fifo_empty            = 0x0040,     // DATA_FIFO_EMPTY
  Status_response_crc_error         = 0x0020,     // CRC_RES_ERR
  Status_read_crc_error             = 0x0010,     // CRC_READ_ERROR
  Status_write_crc_error_no_status  = 0x0008,     // CRC_WRITE_ERROR (2)
  Status_write_crc_error_data       = 0x0004,     // CRC_WRITE_ERROR (1)
  Status_timeout_response           = 0x0002,     // TIME_OUT_RES
  Status_timeout_read               = 0x0001,     // TIME_OUT_READ
};

enum Clock_rate_bits : unsigned
{
  Clock_rate_field_mask             = 0x7,        // CLK_RATE
  Clock_rate_field_shift            = 0,
};

enum Command_data_control_bits : unsigned
{
  // JZ4780/X1600 only.

  Cdc_ccs_expected                  = 0x80000000, // CCS_EXPECTED
  Cdc_read_ce_ata                   = 0x40000000, // READ_CEATA
  Cdc_disable_boot                  = 0x08000000, // DIS_BOOT
  Cdc_expect_boot_ack               = 0x02000000, // EXP_BOOT_ACK
  Cdc_alternative_boot_mode         = 0x01000000, // BOOT_MODE

  // JZ4780 only.

  Cdc_auto_cmd23                    = 0x00040000, // AUTO_CMD23

  // JZ4780/X1600 only.

  Cdc_sdio_interrupt_2cycle         = 0x00020000, // SDIO_PRDT
  Cdc_auto_cmd12                    = 0x00010000, // AUTO_CMD12

  Cdc_recv_fifo_level_field_mask    = 0x3,        // RTRG
  Cdc_fifo_level_16                 = 0,
  Cdc_fifo_level_32                 = 1,
  Cdc_fifo_level_64                 = 2,
  Cdc_fifo_level_96                 = 3,
  Cdc_recv_fifo_level_field_shift   = 14,

  Cdc_trans_fifo_level_field_mask   = 0x3,        // TTRG
  Cdc_trans_fifo_level_field_shift  = 12,

  // Common.

  Cdc_io_abort                      = 0x0800,     // IO_ABORT

  Cdc_bus_width_field_mask          = 0x3,        // BUS_WIDTH
  Cdc_bus_width_field_1bit          = 0,          // = 0
  Cdc_bus_width_field_4bit          = 2,          // = 2
  Cdc_bus_width_field_shift         = 9,

  // JZ4740 only.

  Cdc_dma_enable                    = 0x0100,     // DMA_EN
  Cdc_dma_disable                   = 0x0000,

  // Common.

  Cdc_init_sequence                 = 0x0080,     // INIT

  Cdc_expect_busy                   = 0x0040,     // BUSY
  Cdc_do_not_expect_busy            = 0x0000,

  Cdc_stream_block                  = 0x0020,     // STREAM_BLOCK
  Cdc_not_stream_block              = 0x0000,

  Cdc_write_operation               = 0x0010,     // WRITE_READ
  Cdc_read_operation                = 0x0000,

  Cdc_data_with_command             = 0x0008,     // DATA_EN
  Cdc_no_data_with_command          = 0x0000,

  Cdc_response_format_field_mask    = 0x7,        // RESPONSE_FORMAT
  Cdc_response_format_field_shift   = 0,
};

enum Response_timeout_bits : unsigned
{
  // NOTE: 16-bit value in the JZ4780.
  // NOTE: 32-bit value in the X1600.

  Response_timeout_mask             = 0x000000ff, // RES_TO
};

enum Read_timeout_bits : unsigned
{
  // NOTE: 16-bit value prior to the JZ4780/X1600.

  Read_timeout_mask                 = 0xffffffff, // READ_TO
};

enum Block_length_bits : unsigned
{
  // NOTE: 16-bit value in the JZ4780/X1600.

  Block_length_mask                 = 0x00000fff, // BLK_LEN
};

enum Block_count_bits : unsigned
{
  Block_count_mask                  = 0x0000ffff, // NOB/SNOB
};

// Interrupt mask/flag bits.

enum Interrupt_bits : unsigned
{
  // X1600 only.

  Int_dma_data_done                 = 0x80000000, // DMA_DATA_DONE

  // JZ4780 only.

  Int_auto_cmd23_done               = 0x40000000, // AUTO_CMD23_DONE
  Int_signal_voltage_change         = 0x20000000, // SVS

  // JZ4780/X1600 only.

  Int_pin_level_field_mask          = 0x1f,       // PIN_LEVEL
  Int_pin_level_field_shift         = 24,

  // X1600 only.

  Int_write_request_all_done        = 0x00800000, // WR_ALL_DONE

  // JZ4780/X1600 only.

  Int_boot_crc_error                = 0x00100000, // BCE
  Int_boot_data_end                 = 0x00080000, // BDE
  Int_boot_ack_error                = 0x00040000, // BAE
  Int_boot_ack_received             = 0x00020000, // BAR
  Int_dma_end                       = 0x00010000, // DMAEND
  Int_auto_cmd12_done               = 0x00008000, // AUTO_CMD12_DONE
  Int_data_fifo_full                = 0x00004000, // DATA_FIFO_FULL
  Int_data_fifo_empty               = 0x00002000, // DATA_FIFO_EMP
  Int_crc_response_error            = 0x00001000, // CRC_RES_ERR
  Int_crc_read_error                = 0x00000800, // CRC_READ_ERR
  Int_crc_write_error               = 0x00000400, // CRC_WRITE_ERR
  Int_response_timeout              = 0x00000200, // TIME_OUT_RES
  Int_read_timeout                  = 0x00000100, // TIME_OUT_READ

  // Common.

  Int_sdio                          = 0x80,       // SDIO
  Int_trans_fifo_write_request      = 0x40,       // TXFIFO_WR_REQ
  Int_recv_fifo_read_request        = 0x20,       // RXFIFO_RD_REQ
  Int_end_command_response          = 0x04,       // END_CMD_RES
  Int_programming_done              = 0x02,       // PRG_DONE
  Int_data_transfer_done            = 0x01,       // DATA_TRAN_DONE
};

enum Command_index_bits : unsigned
{
  Command_index_mask                = 0x0000003f, // CMD_INDEX
};

enum Command_argument_bits : unsigned
{
  Command_argument_mask             = 0xffffffff, // ARG
};

enum Response_fifo_bits : unsigned
{
  Response_fifo_mask                = 0x0000ffff, // DATA
};

enum Recv_data_fifo_bits : unsigned
{
  Recv_data_fifo_mask               = 0xffffffff, // DATA
};

enum Trans_data_fifo_bits : unsigned
{
  Trans_data_fifo_mask              = 0xffffffff, // DATA
};

enum Low_power_mode_bits : unsigned
{
  Low_power_mode_enable             = 0x00000001, // LPM
};

enum Dma_control_bits : unsigned
{
  Dma_mode_specify_transfer_length  = 0x80,      // MODE_SEL

  Dma_address_offset_field_mask     = 0x3,        // AOFST
  Dma_address_offset_field_shift    = 5,

  Dma_align_enable                  = 0x10,      // ALIGNEN

  Dma_burst_type_field_mask         = 0x3,        // INCR
  Dma_burst_type_incr16             = 0,
  Dma_burst_type_incr32             = 1,
  Dma_burst_type_incr64             = 2,
  Dma_burst_type_field_shift        = 2,

  Dma_select_common_dma             = 0x02,      // DMASEL
  Dma_select_special_dma            = 0x00,

  Dma_enable                        = 0x01,      // DMAEN
  Dma_disable                       = 0x00,
};



// Command indexes.

enum Command_index : unsigned
{
  Command_go_idle_state             = 0,
  Command_send_op_cond              = 1,
  Command_all_send_cid              = 2,
  Command_send_relative_addr        = 3,          // SD
  Command_set_relative_addr         = 3,          // MMC
  Command_set_dsr                   = 4,
  Command_io_send_op_cond           = 5,          // SDIO
  Command_select_deselect_card      = 7,
  Command_send_if_cond              = 8,
  Command_send_csd                  = 9,
  Command_send_cid                  = 10,
  Command_read_dat_until_stop       = 11,
  Command_stop_transmission         = 12,
  Command_send_status               = 13,
  Command_go_inactive_state         = 15,
  Command_set_blocklen              = 16,
  Command_read_single_block         = 17,
  Command_read_multiple_block       = 18,
  Command_write_dat_until_stop      = 20,
  Command_set_block_count           = 23,
  Command_write_block               = 24,
  Command_write_multiple_block      = 25,
  Command_program_cid               = 26,
  Command_program_csd               = 27,
  Command_set_write_prot            = 28,
  Command_clr_write_prot            = 29,
  Command_send_write_prot           = 30,
  Command_tag_sector_start          = 32,
  Command_tag_sector_end            = 33,
  Command_untag_sector              = 34,
  Command_tag_erase_group_start     = 35,
  Command_tag_erase_group_end       = 36,
  Command_untag_erase_group         = 37,
  Command_erase                     = 38,
  Command_fast_io                   = 39,
  Command_go_irq_state              = 40,
  Command_lock_unlock               = 42,
  Command_io_rw_direct              = 52,         // SDIO
  Command_app_cmd                   = 55,
  Command_gen_cmd                   = 56,
};

// Application-specific command indexes, used by first issuing Command_app_cmd.

enum App_command_index : unsigned
{
  App_command_set_bus_width             = 6,
  App_command_sd_status                 = 13,
  App_command_send_num_wr_blocks        = 22,
  App_command_set_wr_block_erase_count  = 23,
  App_command_sd_send_op_cond           = 41,
  App_command_set_clr_card_detect       = 42,
  App_command_send_scr                  = 51,
  App_command_read_ocr                  = 58,
};

enum Bus_width_bits : unsigned
{
  Bus_width_1bit                        = 0,
  Bus_width_4bit                        = 2,
};

// Command response sizes in 16-bit units.

enum Response_sizes : unsigned
{
  Response_size_R1                      = 3,
  Response_size_R2                      = 8, // omits the CRC and end bit
  Response_size_R3                      = 3,
  Response_size_R4                      = 3,
  Response_size_R5                      = 3,
  Response_size_R6                      = 3,
  Response_size_R7                      = 3,
};

// SD_SEND_OP_COND argument flags.

enum Ocr_argument_flags : unsigned
{
  Ocr_high_capacity_storage             = 0x40000000,
};

// SD_SEND_OP_COND response flags (R3).

enum Ocr_response_flags : unsigned
{
  Ocr_card_powered_up                   = 0x80000000,
};

// R1 status flags.

enum R1_status_flags : unsigned
{
  R1_status_error_mask                  = 0xffff0000,
};



// MMC response structures.
// These are 16-bit aligned to permit conversion to 16-bit arrays.

struct R1
{
  uint8_t end_crc;
  uint32_t status;
  uint8_t index:6, trans_start:2;
} __attribute__((packed,aligned(__alignof__(uint16_t))));

struct R2
{
  // uint8_t end_crc; (not retrieved)

  union
  {
    uint8_t raw[15];
    struct CID cid;
    struct CSD csd;
  } payload;

  uint8_t reserved_trans_start;
} __attribute__((packed,aligned(__alignof__(uint16_t))));

struct R3
{
  uint8_t end_reserved;
  uint32_t ocr;
  uint8_t reserved_trans_start;
} __attribute__((packed,aligned(__alignof__(uint16_t))));

// SDIO response structures.

struct R4
{
  uint8_t end_reserved;
  uint32_t ocr:24, stuff:3, memory_present:1, number_io_functions:3, ready:1;
  uint8_t reserved_trans_start;
} __attribute__((packed,aligned(__alignof__(uint16_t))));

struct R5
{
  uint8_t end_crc;
  uint8_t data;
  uint8_t out_of_range:1, invalid_function_number:1, reserved:1, error:1,
          io_current_state:2, illegal_command:1, crc_error:1;
  uint16_t stuff;
  uint8_t index:6, trans_start:2;
} __attribute__((packed,aligned(__alignof__(uint16_t))));

struct R6
{
  uint8_t end_crc;
  uint16_t status;
  uint16_t rca;
  uint8_t index:6, trans_start:2;
} __attribute__((packed,aligned(__alignof__(uint16_t))));

struct R7
{
  uint8_t end_crc;

  union
  {
    uint32_t check:8, voltage:4, reserved:20;
    uint32_t raw;
  } check_voltage;

  uint8_t index:6, trans_start:2;
} __attribute__((packed,aligned(__alignof__(uint16_t))));
